# pylint: disable=line-too-long,abstract-class-not-used

'''
    Top level function library for ocpkg
'''
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.
import logging
import os
import re
import warnings
import tempfile

import koji
import git
from pyrpkg import Commands, rpkgError
from pyrpkg.utils import cached_property

# doc/ocpkg_man_page.py uses the 'cli' import
from . import cli  # noqa
from .lookaside import StreamLookasideCache, SIGLookasideCache

from pyrpkg.spec import SpecFile
from pyrpkg.utils import is_file_tracked
from ocpkg.sources import SourcesFile

try:
    from rpmautospec import \
        calculate_release_number as rpmautospec_calculate_release_number
    from rpmautospec import process_distgit as rpmautospec_process_distgit
except ImportError:
    rpmautospec_process_distgit = None
    rpmautospec_calculate_release_number = None


_DEFAULT_VERSION = '8'


class DistGitDirectory(object):

    signame = None
    opencloudosversion = _DEFAULT_VERSION
    projectname = None
    releasename = None
    distrobranch = False
    repo = None
    git_origin_substr = 'git@gitee.com/src-opencloudos-rpms/'

    def __init__(self, branchtext, repo_path=None):
        if repo_path:
            # self.repo = git.cmd.Git(repo_path)
            self.repo = git.repo.Repo(repo_path)
        rhelbranchre = r'rhel-(?P<major>\d+)\.(?P<minor>\d+)(?:\.(?P<appstream>\d+))?'
        sigtobranchre = r'oc(?P<opencloudosversion>\d+)-?(?P<projectname>.+)?'
        distrobranchre = r'c(?P<opencloudosversion>\d+)-?(?P<projectname>\w+)?'
        oldbranchre = r'(?P<signame>\w+)(?P<opencloudosversion>\d)'
        rhelmatch = re.search(rhelbranchre, branchtext)
        sigmatch = re.match(sigtobranchre, branchtext)
        distromatch = re.match(distrobranchre, branchtext)
        oldbranchmatch = re.match(oldbranchre, branchtext)
        if rhelmatch:
            gd = rhelmatch.groupdict()
            self.distrobranch = True
            self.signame = 'opencloudos'
            self.opencloudosversion = gd['major']
        elif sigmatch:
            gd = sigmatch.groupdict()
            # self.signame = gd['signame']
            self.opencloudosversion = gd['opencloudosversion']
            self.projectname = gd['projectname']
            self.releasename = 'oc' + gd['opencloudosversion']
            self.signame = ''
            # Users have the option to specify (or not specify) common in their
            # git repos. Ww need to handle these cases because common is not a
            # project nor is it a release.
            # if gd['projectname'] != 'common':
            #     self.projectname = gd['projectname']
            # if gd['releasename'] != 'common':
            #     self.releasename = gd['releasename']
        elif distromatch:
            gd = distromatch.groupdict()
            self.distrobranch = True
            self.signame = 'opencloudos'
            self.opencloudosversion = gd['opencloudosversion']

            if gd['projectname'] != 'common':
                self.projectname = gd['projectname']
        elif oldbranchmatch:
            warnings.warn("This branch is deprecated and will be removed soon",
                          DeprecationWarning)
        else:
            if not self.is_fork():
                raise ValueError("Branchname: {0} is not valid".format(branchtext))
            else:
                self.distrobranch = True
                self.signame = 'opencloudos'
                self.projectname = self.get_origin().split('_')[-1].replace('.git', '')

                warnings.warn('Remote "origin" was detected as a fork, ignoring branch name checking')

    def get_origin(self):
        if self.repo is None:
            return ''
        if 'origin' not in self.repo.remotes:
            return ''
        urls = [u for u in self.repo.remotes['origin'].urls]
        if len(urls) == 0:
            return ''
        return urls[0]

    def is_fork(self):
        """
        Check if origin remote repository is using a fork url.

        Returns
        bool
            A boolean flag indicating if origin remote url is using
            a forked repository url.
        """
        # git+ssh://git@gitlab.com/redhat/opencloudos-stream/rpms/binutils.git
        if self.repo is None:
            return False
        return self.git_origin_substr not in self.get_origin()



class Commands(Commands):
    '''
        For the pyrpkg commands with ocpkg behavior
    '''
    def __init__(self, *args, **kwargs):
        '''
            Init the object and some configuration details.
        '''
        super(Commands, self).__init__(*args, **kwargs)
        # For MD5 we want to use the old format of source files, the BSD format
        # should only be used when configured for SHA512
        self.source_entry_type = 'bsd' if self.lookasidehash not in ('md5', 'sha1') else 'old'
        self.branchre = 'oc\d{1,}.+|master'
        self.lookaside_structure = 'branch'

    @property
    def distgitdir(self):
        return DistGitDirectory(self.branch_merge, repo_path=self.path)

    @cached_property
    def lookasidecache(self):
        return SIGLookasideCache(self.lookasidehash,
                                 self.lookaside,
                                 self.lookaside_cgi,
                                 self.repo_name,
                                 self.branch_merge,
                                 self.lookaside_structure
                                 )

    # redefined loaders
    def load_rpmdefines(self):
        '''
            Populate rpmdefines based on branch data
        '''

        if not self.distgitdir.opencloudosversion:
            raise rpkgError('Could not get the OS version from the branch:{0}'.format(self.branch_merge))

        self._distvar = self.distgitdir.opencloudosversion
        self._distval = self._distvar.replace('.', '_')

        self._disttag = 'oc%s' % self._distval
        self._rpmdefines = ["--define", "_sourcedir %s" % self.layout.sourcedir,
                            "--define", "_specdir %s" % self.layout.specdir,
                            "--define", "_builddir %s" % self.layout.builddir,
                            "--define", "_srcrpmdir %s" % self.layout.srcrpmdir,
                            "--define", "_rpmdir %s" % self.layout.rpmdir,
                            "--define", "dist .%s" % self._disttag,
                            # int and float this to remove the decimal
                            "--define", "%s 1" % self._disttag]
        self.log.debug("RPMDefines: %s" % self._rpmdefines)

    def construct_build_url(self, *args, **kwargs):
        """Override build URL for OpenCloudOS Koji build

        In OpenCloudOS Koji, anonymous URL should have prefix "git+https://"
        """
        url = super(Commands, self).construct_build_url(*args, **kwargs)
        return 'git+{0}'.format(url)
    
    @property
    def target(self):
        """This property ensures the target attribute"""

        if not self._target:
            self.load_target()
        return self._target

    @target.setter
    def target(self, value):
        self._target = value

    def load_target(self):
        """ This sets the target attribute (used for mock and koji) """

        self._target = "dist-{}".format(self.distgitdir.releasename)

    @property
    def sources_filename(self):
        return os.path.join(self.path, '.{}.metadata'.format(self.repo_name))

    def sources(self, outdir=None):
        """Download source files"""

        if not os.path.exists(self.sources_filename):
            self.log.info("sources file doesn't exist. Source files download skipped.")
            return

        # Default to putting the files where the repository is
        if not outdir:
            outdir = self.path

        sourcesf = SourcesFile(self.sources_filename, self.source_entry_type)

        try:
            specf = SpecFile(os.path.join(self.layout.specdir, self.spec),
                             self.layout.sourcedir)
            spec_parsed = True
        except Exception:
            self.log.warn("Parsing specfile for used sources failed. "
                          "Falling back to downloading all sources.")
            spec_parsed = False

        args = dict()
        if self.lookaside_request_params:
            if 'branch' in self.lookaside_request_params.split():
                # The value of branch_merge is dynamic property;  to get it's
                # value you need to be in proper branch or you need to first
                # specify --release (which is pretty annoying).  Since not every
                # dist-git instance out there really needs 'branch' argument to
                # expand lookaside cache urls - make it optional.
                args['branch'] = self.branch_merge

        for entry in sourcesf.entries:
            outfile = os.path.join(outdir, entry.file)
            if is_file_tracked(outfile, outdir):
                raise rpkgError(
                    "Error: Attempting a download '{0}' that would override a git tracked file. "
                    "Either remove the corresponding line from 'sources' file to keep the git "
                    "tracked one or 'git rm' the file to allow the download.".format(outfile))
            if (spec_parsed and os.path.basename(entry.file) not in specf.sources):
                self.log.info("Not downloading unused %s" % entry.file)
                continue
            # import pdb;            pdb.set_trace()
            try:
                self.lookasidecache.download(
                    self.ns_repo_name if self.lookaside_namespaced else self.repo_name,
                    entry.file, entry.hash, outfile,
                    hashtype=entry.hashtype, **args
                )
            except:
                logging.exception("")

    def upload(self, files, replace=False, offline=False):
        """Upload source file(s) in the lookaside cache

        Both file `sources` and `.gitignore` will be updated with uploaded
        files, and added to index tree eventually.

        :param iterable files: an iterable of files to upload.
        :param bool replace: optionally replace the existing tracked sources.
            Defaults to `False`.
        :param bool offline: do all the steps except uploading into lookaside cache
        :raises rpkgError: if failed to add a file to file `sources`.
        """
        from pyrpkg.gitignore import GitIgnore
        from pyrpkg.errors import (AlreadyUploadedError, HashtypeMixingError,
                                   UnknownTargetError, rpkgAuthError, rpkgError)
        import sys

        sourcesf = SourcesFile(self.sources_filename, self.source_entry_type,
                               replace=replace)
        gitignore = GitIgnore(os.path.join(self.path, '.gitignore'))
        all_files_already_uploaded = True

        for f in files:
            # TODO: Skip empty file needed?
            file_hash = self.lookasidecache.hash_file(f)
            # file_basename = os.path.basename(f)

            try:
                sourcesf.add_entry(self.lookasidehash, f, file_hash)
            except HashtypeMixingError as e:
                msg = '\n'.join([
                    'Can not upload a new source file with a %(newhash)s '
                    'hash, as the "%(sources)s" file contains at least one '
                    'line with a %(existinghash)s hash.', '',
                    'Please redo the whole "%(sources)s" file using:',
                    '    `%(arg0)s new-sources file1 file2 ...`']) % {
                        'newhash': e.new_hashtype,
                        'existinghash': e.existing_hashtype,
                        'sources': self.sources_filename,
                        'arg0': sys.argv[0],
                    }
                raise rpkgError(msg)

            gitignore.add('%s' % f)
            if is_file_tracked(f, self.path):  # need full file path
                raise rpkgError(
                    "Error: Attempting to upload a git tracked file '{0}'. Only upload files not "
                    "tracked by git. You can use 'git rm --cached' to untrack a file from git.\n"
                    "Hint: Use git for text files like the spec file, patches or helper scripts. "
                    "Use the lookaside cache for binary blobs, usually upstream source "
                    "tarballs.".format(f))
            try:
                self.lookasidecache.upload(
                    self.ns_repo_name if self.lookaside_namespaced else self.repo_name,
                    f, file_hash, offline=offline)
            except AlreadyUploadedError:
                pass
            else:
                all_files_already_uploaded = False

        sourcesf.write()
        gitignore.write()

        self.repo.index.add([self.sources_filename, '.gitignore'])

        if all_files_already_uploaded:
            raise AlreadyUploadedError('File already uploaded')

    def mockbuild(self, mockargs=None, root=None, hashtype=None, shell=None,
                  force_local_mock_config=None, srpm_mock=False):
        """Build the package in mock, using mockargs

        Log the output and returns nothing

        :param list mockargs: list of command line arguments which are passed
            mock.
        :param str root: chroot config name which is passed to mock ``-r``
            option.
        :param str hashtype: used to generate SRPM only if there is no SRPM
            generated before.
        :param bool shell: indicate whether to go into chroot.
        :param bool force_local_mock_config: enforce download of the Mock
            configuration from Kojihub (True), or enforce local Mock config
            (False). If None local configuration is used if present.
        :param bool srpm_mock: genereate source rpm

        .. versionadded:: 1.56
           Parameter shell.
        .. versionadded:: 1.64
           Parameter srpm_mock.
        """
        if mockargs is None:
            mockargs = []
        # Make sure we have an srpm to run on if we need it
        if srpm_mock or (hasattr(self, 'srpmname') and self.srpmname):
            # srpm is being generated or mockbuild is running with path
            # to already generated srpm
            pass
        else:
            self.srpm(hashtype=hashtype)

        # setup the command
        cmd = ['mock']
        cmd.extend(mockargs)
        if self.quiet:
            cmd.append('--quiet')

        root, config_dir = self.get_mock_config_dir(root, force_local_mock_config)
        if config_dir:
            cmd.extend(['--configdir', config_dir])

        cmd += ['-r', root, '--resultdir', self.mock_results_dir]

        if shell:
            cmd.append('--shell')
        elif srpm_mock:
            cmd += ['--buildsrpm', '--sources', self.layout.sourcedir, '--spec', self.spec]
        else:
            cmd += ['--rebuild', self.srpmname]

        # Run the command
        try:
            self._run_command(cmd)
        finally:
            self.log.debug('Cleaning up mock temporary config directory: %s', config_dir)
            self._cleanup_tmp_dir(config_dir)

    def srpm(self, hashtype=None, define=None, builddir=None, buildrootdir=None,
             arch=None, extra_args=None):
        """Create an srpm using hashtype from content

        Requires sources already downloaded. The generated SRPM file will be
        put into package repository directory.

        :param str hashtype: an alternative algorithm used for payload file
            digests.
        :param list define: an optional list of rpmbuild macros.
        :param str builddir: optionally define an alternate builddir.
        :param str buildrootdir: optionally define an alternate buildrootdir.
        :param str arch: optional to run prep section for a specific arch. By
            default, local system arch will be used.
        :param list extra_args: additional arguments that are passed to
            the rpmbuild command.
        """
        self.srpmname = os.path.join(self.layout.srcrpmdir,
                                     "%s-%s-%s.src.rpm"
                                     % (self.repo_name, self.ver, self.rel))

        # See if we need to build the srpm
        if os.path.exists(self.srpmname):
            self.log.debug('Srpm found, rewriting it.')

        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)
        if builddir:
            # Tack on a new builddir to the end of the defines
            cmd.append("--define '_builddir %s'" % os.path.abspath(builddir))
        if arch:
            cmd.extend(['--target', arch])
        if define:
            for entry in define:
                cmd.extend(['--define', entry])
        if extra_args:
            cmd.extend(extra_args)
            self.log.debug("Extra args '{0}' are passed to rpmbuild "
                           "command".format(extra_args))
        if self.quiet:
            cmd.append('--quiet')
        if buildrootdir:
            cmd.append("--define '_buildrootdir {0}'".format(
                os.path.abspath(buildrootdir)))

        # Figure out which hashtype to use, if not provided one
        if not hashtype:
            # Try to determine the dist
            hashtype = self._guess_hashtype()
        # This may need to get updated if we ever change our checksum default
        if not hashtype == 'sha256':
            cmd.extend(["--define '_source_filedigest_algorithm %s'"
                        % hashtype,
                        "--define '_binary_filedigest_algorithm %s'"
                        % hashtype])
        specpath = os.path.join(self.layout.specdir, self.spec)
        tmpdir = None
        try:
            if not self.uses_rpmautospec or not rpmautospec_process_distgit:
                cmd.extend(['--nodeps', '-bs', specpath])
            else:
                tmpdir = tempfile.mkdtemp(prefix="rpkg-rpmautospec")
                tmpspecpath = os.path.join(tmpdir, self.spec)
                rpmautospec_process_distgit(specpath, tmpspecpath)
                cmd.extend(['--nodeps', '-bs', tmpspecpath])
            self._run_command(cmd)
        finally:
            self._cleanup_tmp_dir(tmpdir)
